package dev.fyloz.trial.colorrecipesexplorer

import dev.fyloz.colorrecipesexplorer.databasemanager.CreDatabase
import dev.fyloz.colorrecipesexplorer.databasemanager.databaseContext
import dev.fyloz.colorrecipesexplorer.databasemanager.databaseUpdaterProperties
import org.slf4j.Logger
import org.springframework.boot.context.properties.ConfigurationProperties
import org.springframework.boot.jdbc.DataSourceBuilder
import org.springframework.context.annotation.Bean
import org.springframework.context.annotation.Configuration
import org.springframework.core.env.Environment
import javax.sql.DataSource

const val SUPPORTED_DATABASE_VERSION = 2
const val ENV_VAR_ENABLE_DATABASE_UPDATE_NAME = "CRE_ENABLE_DB_UPDATE"
val DATABASE_NAME_REGEX = Regex("(\\w+)$")

@Configuration
class DataSourceConfiguration {
    @Bean(name = ["dataSource"])
    @ConfigurationProperties(prefix = "spring.datasource")
    fun customDataSource(
        logger: Logger,
        environment: Environment,
        databaseUpdaterProperties: DatabaseUpdaterProperties
    ): DataSource {
        val databaseUrl: String = environment.getProperty("spring.datasource.url")!!

        runDatabaseVersionCheck(logger, databaseUrl, databaseUpdaterProperties)

        return DataSourceBuilder
            .create()
            .url(databaseUrl)  // Hikari won't start without that
            .build()
    }
}

/**
 * Runs a database version check. If the database's version is not supported and the environment variable 'CRE_ENABLE_DB_UPDATE' is:
 * - 1: The database will be updated.
 * - Any other value: The application will not start.
 */
fun runDatabaseVersionCheck(logger: Logger, databaseUrl: String, databaseUpdaterProperties: DatabaseUpdaterProperties) {
    logger.info("Verifying database's version before starting...")

    if (":h2:" in databaseUrl) {
        logger.warn("H2 is not supported by the database manager")
        return
    }

    val databaseUpdateEnabled = System.getenv(ENV_VAR_ENABLE_DATABASE_UPDATE_NAME) == "1"
    val database = getDatabase(databaseUrl, databaseUpdaterProperties, logger)
    val version = database.version

    if (version != SUPPORTED_DATABASE_VERSION) {
        if (!databaseUpdateEnabled) {
            database.close()
            throwUnsupportedDatabaseVersion(version, logger)
        } else runDatabaseUpdate(logger, database)
    } else {
        logger.info("The database is up to date!")

        if (databaseUpdateEnabled) {
            logger.warn("The database is up to date but the environment variable '$ENV_VAR_ENABLE_DATABASE_UPDATE_NAME' is set to '1'. Set it to '0' to prevent any possible data loss.")
        }
    }

    database.close()
}

/** Updates the given [database]. */
fun runDatabaseUpdate(logger: Logger, database: CreDatabase) {
    logger.info("The environment variable '$ENV_VAR_ENABLE_DATABASE_UPDATE_NAME' is set to '1'; The database will be updated.")
    database.update()
}

fun getDatabase(
    databaseUrl: String,
    databaseUpdaterProperties: DatabaseUpdaterProperties,
    logger: Logger
): CreDatabase {
    val databaseName =
        (DATABASE_NAME_REGEX.find(databaseUrl) ?: throw DatabaseVersioningException.InvalidUrl(databaseUrl)).value

    return CreDatabase(
        databaseContext(
            properties = databaseUpdaterProperties(
                targetVersion = SUPPORTED_DATABASE_VERSION,
                url = databaseUrl.removeSuffix(databaseName),
                dbName = databaseName,
                username = databaseUpdaterProperties.username,
                password = databaseUpdaterProperties.password
            ),
            logger
        )
    )
}

fun throwUnsupportedDatabaseVersion(version: Int, logger: Logger) {
    if (version > SUPPORTED_DATABASE_VERSION) {
        logger.error("Version $version of the database is not supported; Only version $SUPPORTED_DATABASE_VERSION is currently supported; Update this application to use the database.")
    } else {
        logger.error(
            """Version $version of the database is not supported; Only version $SUPPORTED_DATABASE_VERSION is currently supported.
            |You can update the database to the supported version by either:
            |	- Setting the environment variable '$ENV_VAR_ENABLE_DATABASE_UPDATE_NAME' to '1' to update the database automatically
            |	- Updating the database manually with the database manager utility (https://git.fyloz.dev/color-recipes-explorer/database-manager)
            |	
            |Don't forget to backup the database before upgrading to prevent any data loss.
        """.trimMargin()
        )
    }
    throw DatabaseVersioningException.UnsupportedDatabaseVersion(version)
}

@ConfigurationProperties(prefix = "databaseupdater")
class DatabaseUpdaterProperties {
    var username: String = ""
    var password: String = ""
}

sealed class DatabaseVersioningException(message: String) : Exception(message) {
    class InvalidUrl(url: String) : DatabaseVersioningException("Invalid database url: $url")
    class UnsupportedDatabaseVersion(version: Int) :
        DatabaseVersioningException("Unsupported database version: $version; Only version $SUPPORTED_DATABASE_VERSION is currently supported")
}
